#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#define NUM_FUNC                    34
#define MAX_BYTES                   32

typedef struct tagFLINE             FLINE, *PFLINE;

struct tagFLINE
{
  PFLINE                      pPrev;
  PFLINE                      pNext;

  unsigned long               uAddr;
  unsigned long               uBytes;
  unsigned long               uOprAddr;
  char                        szMnemonic[64];
  char                        szOperands[128];
  char                        szLabelRef[64];
  unsigned char               byOpc[MAX_BYTES];
};

int InsertMnemonic ( PFLINE pIns, char *pszMne, char *pszOpr )
{
  PFLINE        pFL = (PFLINE)malloc(sizeof(FLINE));

  if (!pFL) return 1;
  memset(pFL,0,sizeof(FLINE));

  strcpy(pFL->szMnemonic,pszMne);
  strcpy(pFL->szOperands,pszOpr);

  pFL->pPrev = pIns->pPrev;
  pFL->pNext = pIns;

  pIns->pPrev = pFL;

  pFL->pPrev->pNext = pFL;

  return 0;
}


char *IgnWS ( char *p )
{
  while ((*p) && (((unsigned char)(*p))<0x21)) p++;
  return p;
}

char *IgnNWS ( char *p )
{
  while ((*p) && (((unsigned char)(*p))>0x20)) p++;
  return p;
}

PFLINE pFunctions[NUM_FUNC];

typedef struct tagFCT  FCT, *PFCT;

struct tagFCT
{
  char          szFunction[64];
  int           grep;
  int           align;
};

FCT Functions[NUM_FUNC] =
{
/* 0*/  { "InterlockedCompareExchangePointer"   , 1,  5 },    // *
/* 1*/  { "_PGOPTI_Prof_Dump"                   , 0,  0 },    // *
/* 2*/  { "_PGOPTI_Prof_Reset"                  , 0,  0 },    // *
/* 3*/  { "_PGOPTI_Prof_Dump_And_Reset"         , 0,  0 },    // *
/* 4*/  { "_PGOPTI_Set_Interval_Prof_Dump"      , 0,  0 },    // *
/* 5*/  { "_PGOPTI_Prof_Begin"                  , 1, -5 },    // *
/* 6*/  { "_PGOPTI_Prof_Count"                  , 1, -5 },    // *
/* 7*/  { "_PGOPTI_Prof_If"                     , 1, -5 },    // *
/* 8*/  { "_PGOPTI_Prof_Switch"                 , 1, -5 },    // *
/* 9*/  { "_PGOPTI_Prof_Case"                   , 1, -5 },    // *
/*10*/  { "_PGOPTI_Prof_Arithmetic_If"          , 1, -5 },    // *
/*11*/  { "_PGOPTI_Prof_Computed_Goto"          , 1, -5 },    // *
/*12*/  { "_PGOPTI_Prof_Igoto"                  , 1, -5 },    // *
/*13*/  { "_PGOPTI_Prof_Tag"                    , 1, -5 },    // *
/*14*/  { "_PGOPTI_Prof_Icall"                  , 1, -5 },    // *
/*15*/  { "_PGOPTI_Prof_Div_64_VP"              , 1, -5 },    // *
/*16*/  { "_PGOPTI_Prof_Div_VP"                 , 1, -5 },    // *
/*17*/  { "_PGOPTI_Prof_Th_VP"                  , 1, -5 },    // *
/*18*/  { "pgopti_itarg_hash"                   , 1,  2 },    // *
/*19*/  { "pgopti_Set_Interval_Prof_Dump"       , 0,  0 },    // *
/*20*/  { "pgopti_Prof_Dump"                    , 0,  0 },    // *
/*21*/  { "pgopti_Prof_Dump_And_Reset"          , 0,  0 },    // *
/*22*/  { "pgopti_Reset_Dynamic_Info"           , 1,  2 },    // *
/*23*/  { "pgopti_Dump_Segment"                 , 1,  2 },    // *
/*24*/  { "pgopti_Write_Disk"                   , 0,  0 },    // *
/*25*/  { "pgopti_Disk_Alloc"                   , 0,  0 },    // *
/*26*/  { "pgopti_Sizeof_Segment"               , 1,  2 },    // *
/*27*/  { "pgopti_Prof_Dump_In_Progress"        , 1,  5 },    // *
/*28*/  { "pgopti_Prof_Dump_Exit"               , 1,  5 },    // *
/*29*/  { "pgopti_Value_Prof_LFE_Insert"        , 1,  2 },    // *
/*30*/  { "pgopti_Value_Prof_Sort"              , 1,  2 },    // *
/*31*/  { "pgopti_Value_Prof_Clear_Non_Steady"  , 1,  2 },    // *
/*32*/  { "pgopti_Check_Prof_Dump_Interval"     , 0,  0 },    // *
/*33*/  { "pgopti_Value_Max"                    , 1,  2 }     // *
};

#define ISHEX(c)        ((((c)>='0') && ((c)<='9')) || (((c)>='A') && ((c)<='F')) || (((c)>='a') && ((c)<='f')))

FILE                 *hFile;
char                  szLine[1024];
char                  szCopyright[1024];

int                   labels = 0;
unsigned long         uLabAddrs[1024];
int                   iLabRefs[1024];

unsigned long         addr_is64bit     = 0;
unsigned long         addr_uValue1     = 0;
unsigned long         addr_uValue2     = 0;
unsigned long         addr_pSegments   = 0;
unsigned long         addr_uFileOffset = 0;
unsigned long         addr_uMaxVal     = 0;
unsigned long         addr_uLock       = 0;

int CreateLabel ( char *pszLabel, unsigned long uAddr )
{
  int i;

  if (!labels)
  {
    uLabAddrs[labels] = uAddr;
    sprintf(pszLabel,"_L%03lX",labels++);
    return labels-1;
  }

  if (labels==1024) return -1;

  for (i=0;i<labels;i++)
    if (uLabAddrs[i]==uAddr)
    {
      sprintf(pszLabel,"_L%03lX",(unsigned long)i);
      return i;
    }

  uLabAddrs[labels] = uAddr;
  sprintf(pszLabel,"_L%03lX",labels++);
  return labels-1;
}

int GetLabel ( char *pszLabel, unsigned long uAddr )
{
  int i;

  if (!labels)
  {
    uLabAddrs[labels] = uAddr;
    iLabRefs[labels]++;
    sprintf(pszLabel,"_L%03lX",labels++);
    return 0;
  }

  if (labels==1024) return 1;

  for (i=0;i<labels;i++)
    if (uLabAddrs[i]==uAddr)
    {
      iLabRefs[i]++;
      sprintf(pszLabel,"_L%03lX",(unsigned long)i);
      return 0;
    }

  uLabAddrs[labels] = uAddr;
  iLabRefs[labels]++;
  sprintf(pszLabel,"_L%03lX",labels++);
  return 0;
}

//  1 = error
//  0 = okay
// -1 = EOF (end of file/function)

int GrepLine ( PFLINE *ppRoot, PFLINE *ppAct )
{
  PFLINE                        pFL = (PFLINE)malloc(sizeof(FLINE));
  char                         *p,*p2,*p3;
  PFLINE                        pAct;
  int                           l;

  if (!pFL) return 1;

  NextLine:
  memset(pFL,0,sizeof(FLINE));

  szLine[0]=0;
  if (!fgets(szLine,sizeof(szLine),hFile))
  {
    free(pFL);
    return -1;
  }

  p = IgnWS(szLine);
  if (!(*p))
  {
    free(pFL);
    return -1;
  }

  pFL->uAddr = strtoul(p,&p,16);
  if (!pFL->uAddr)
  {
    ErrExit:
    free(pFL);
    return 1;
  }

  if ((!p) || ((*p)!=':')) goto ErrExit;

  p=IgnWS(p+1);

  while ((ISHEX(p[0])) && (ISHEX(p[1])) && (((unsigned char)p[2])<0x21)) 
  {
    if (pFL->uBytes==MAX_BYTES) goto ErrExit;
    pFL->byOpc[pFL->uBytes++] = (unsigned char)strtoul(p,&p,16);
    p=IgnWS(p);
  }

  if (!(*p))  // concat opcode bytes (previous statement)
  {
    pAct = *ppAct;
    if (!pAct) goto ErrExit;

    if ((pAct->uBytes+pFL->uBytes)>MAX_BYTES) goto ErrExit;

    memcpy(&pAct->byOpc[pAct->uBytes],pFL->byOpc,pFL->uBytes);
    pAct->uBytes += pFL->uBytes;
    free(pFL);
    return 0;
  }

  p2 = IgnNWS(p);
  l  = p2-p;
  if (l<=0) goto ErrExit;

  memcpy(pFL->szMnemonic,p,l);
  pFL->szMnemonic[l]=0x00;

  if ( (!strcmp(pFL->szMnemonic,"lock")) ||
       (!strcmp(pFL->szMnemonic,"repz")) ||
       (!strcmp(pFL->szMnemonic,"repnz")) ||
       (!strcmp(pFL->szMnemonic,"repe")) ||
       (!strcmp(pFL->szMnemonic,"repne")) )
  {
    p = IgnWS(p2);
    p2 = IgnNWS(p);
    l  = p2-p;
    if (l<=0) goto ErrExit;
    *p2=0;
    strcat(pFL->szMnemonic,"; ");
    strcat(pFL->szMnemonic,p);
    p = IgnWS(p2+1);
  }
  else
    p = IgnWS(p2);

  l=strlen(p);
  if ((l) && (p[l-1]=='\n')) p[l-1]=0;

  strcpy(pFL->szOperands,p);

  p = strchr(pFL->szOperands,'<');
  if (p)
  {
    p2 = strchr(p+1,'>');
    if (p2)
    {
      pFL->uOprAddr = strtoul(pFL->szOperands,&p3,16);
      if ((pFL->uOprAddr) && (p3) && (p3[0]==' ') && (p==(p3+1)))
      {
        if ((strchr(p,'+')) || (strchr(p,'-'))) // label
        {
          if (GetLabel(pFL->szOperands,pFL->uOprAddr))
            goto ErrExit;
        }
        else
        {
          *p2 = 0;
          strcpy(pFL->szLabelRef,p+1);
          strcpy(pFL->szOperands,pFL->szLabelRef);
        }
      }
    }
  }

  if ((!strcmp(pFL->szMnemonic,"call")) && (!strcmp(pFL->szOperands,"pgopti_Check_Prof_Dump_Interval")))
    pFL->szMnemonic[0]=0;    
  else
  if ((!strcmp(pFL->szMnemonic,"nop")) && (!pFL->szOperands[0]))
    pFL->szMnemonic[0]=0;    
  else
  if ((!strcmp(pFL->szMnemonic,"cltd")) && (!pFL->szOperands[0]))
  {
    strcpy(pFL->szMnemonic,".byte");
    strcpy(pFL->szOperands,"0x99");
  }

  pAct = *ppAct;
  if (pAct)
  {
    pAct->pNext = pFL;
    pFL->pPrev  = pAct;
    *ppAct      = pFL;
  }
  else
  {
    *ppRoot = *ppAct = pFL;
  }

  return 0;
}

int ProcessFunction ( char *pszFunction )
{
  int                 i,j,iLIdx,cnt;
  PFLINE              pRoot=NULL,pAct=NULL,pSave,pRun,pMod;
  char               *p,szLabel[256];
  unsigned long       uTestAddr,uPatchAddr;
  char                szOpr[256];

  for (i=0;i<NUM_FUNC;i++)
    if (!strcmp(pszFunction,Functions[i].szFunction)) break;

  if (i==NUM_FUNC)
  {
    szLine[0]=0;
    while (fgets(szLine,sizeof(szLine),hFile))
    {
      if ((szLine[0]=='\n') && (!szLine[1]))
        return 0;
      szLine[0]=0;
    }
    return 0;
  }

  fprintf(stdout,"Extracting function '%s'... ",Functions[i].szFunction);
  fflush(stdout);

  do
  {
    j = GrepLine(&pRoot,&pAct);
    if (j==1) { fprintf(stdout,"ERROR.\n");return 1; }
  }
  while (!j);   

  if (!pRoot) { fprintf(stdout,"ERROR.\n");return 1; }

  pFunctions[i] = pRoot;
  
  fprintf(stdout,"DONE.\n");

  while (pAct)
  {
    if ( ((pAct->uBytes==1) && (pAct->byOpc[0]==0x90)) ||
         ((pAct->uBytes==2) && (pAct->byOpc[0]==0x89) && (pAct->byOpc[1]==0xF6)) ||
         ((pAct->uBytes==3) && (pAct->byOpc[0]==0x8D) && (pAct->byOpc[1]==0x76) && (pAct->byOpc[2]==0x00)) )
    {
      pAct->szMnemonic[0]=0;
      pAct = pAct->pPrev;
    }
    else
      break;
  }

  pMod=pAct;
  while (pMod)
  {
    if (( ((pMod->uBytes==1) && (pMod->byOpc[0]==0x90)) ||
          ((pMod->uBytes==2) && (pMod->byOpc[0]==0x89) && (pMod->byOpc[1]==0xF6)) ||
          ((pMod->uBytes==3) && (pMod->byOpc[0]==0x8D) && (pMod->byOpc[1]==0x76) && (pMod->byOpc[2]==0x00)) ) &&
           (pMod->pPrev) && (!strcmp(pMod->pPrev->szMnemonic,"jmp")) )
    {
      pMod->szMnemonic[0]=0;
    }
    pMod = pMod->pPrev;
  }

  switch(i)
  {
    case 5: // _PGOPTI_Prof_Begin
      while (pAct)
      {
        if (!strcmp(pAct->szMnemonic,"ret")) break;
        pAct->szMnemonic[0]=0;
        pAct=pAct->pPrev;
      }
      if (!pAct)
      {
        AbortPP:
        fprintf(stderr,"ERROR (post processing, patching).\n");
        return 1;
      }
      pAct=pAct->pPrev;
      if ( (strcmp(pAct->szMnemonic,"pop")) || (strcmp(pAct->szOperands,"%ebx")) ) goto AbortPP;
      pAct=pAct->pPrev;if (!pAct) goto AbortPP;
      
      while (pAct)
      {
        if ( (!strcmp(pAct->szMnemonic,"movl")) && ((p=strstr(pAct->szOperands,"$0x1,"))!=NULL) )
        {
          addr_is64bit = strtoul(p+5,&p,0);
          if (!addr_is64bit) goto AbortPP;
          pAct=pAct->pPrev;
          break;
        }
        else
          pAct->szMnemonic[0]=0;
        pAct = pAct->pPrev;
      }
      if (!pAct) goto AbortPP;

      while (pAct)
      {
        if ( (!strcmp(pAct->szMnemonic,"mov")) && ((p=strstr(pAct->szOperands,"%edx,"))!=NULL) )
        {
          addr_uValue2 = strtoul(p+5,&p,0);
          if (!addr_uValue2) goto AbortPP;
          pAct=pAct->pPrev;

          if ( (!strcmp(pAct->szMnemonic,"mov")) && ((p=strstr(pAct->szOperands,"%eax,"))!=NULL) )
          {
            addr_uValue1 = strtoul(p+5,&p,0);
            if (!addr_uValue1) goto AbortPP;
          }
          else goto AbortPP;
          pAct = pAct->pPrev;
          break;
        }
        pAct = pAct->pPrev;
      }
      if (!pAct) goto AbortPP;

      while (pAct)
      {
        if ( (!strcmp(pAct->szMnemonic,"call")) && (!strcmp(pAct->szOperands,"InterlockedCompareExchangePointer")) )
        {
          pAct = pAct->pPrev;
          if (!pAct) goto AbortPP;
          if ( (!strcmp(pAct->szMnemonic,"push")) && (pAct->szOperands[0]=='$') && 
               (pAct->szOperands[1]=='0') && (pAct->szOperands[2]=='x') 
             )
          addr_pSegments = strtoul(&pAct->szOperands[1],&p,0);
          if (!addr_pSegments) goto AbortPP;
          pAct=pAct->pPrev;
          break;
        }
        pAct = pAct->pPrev;
      }
      if (!pAct) goto AbortPP;
      break;

    case 6:   // _PGOPTI_Prof_Count  (delete __assert)
    case 7:   // 
    case 9:   //
    case 10:  //
    case 11:  //
    case 13:  //
    case 16:  //
    case 17:  //
      while (pAct)
      {
        if ((!strcmp(pAct->szMnemonic,"add")) && (!strcmp(pAct->szOperands,"$0x10,%esp"))) break;
        pAct=pAct->pPrev;
      }
      while (pAct)
      {
        if (!strcmp(pAct->szMnemonic,"ret")) break;
        pAct->szMnemonic[0]=0;
        pAct=pAct->pPrev;
      }
      if (!pAct) goto AbortPP;
      pSave = pAct->pNext;
      if (!pSave) goto AbortPP;
      
      pAct=pAct->pPrev;
      if (!pAct) goto AbortPP;
      while (pAct)
      {
        if (strcmp(pAct->szMnemonic,"pop")) break;
        pAct=pAct->pPrev;
      }
      if (!pAct) goto AbortPP;
      pAct = pAct->pNext;
      if (!pAct) goto AbortPP;
      
      uPatchAddr = pAct->uAddr;
      if (-1==(iLIdx=CreateLabel(szLabel,uPatchAddr)))
      {
        fprintf(stdout,"ERROR (cannot create patch label).\n");
        return 1;
      }
      pRun = pRoot;
      while (pRun!=pAct)
      {
        if ((pRun->uOprAddr>=pSave->uAddr) && (!strcmp(pRun->szMnemonic,"jae")))
        {
          strcpy(pRun->szLabelRef,szLabel);
          strcpy(pRun->szOperands,pRun->szLabelRef);
          iLabRefs[iLIdx]++;
        }
        pRun=pRun->pNext;
      }
      break;

    case 22:  // pgopti_Reset_Dynamic_Info
      pAct = pRoot;
      while (pAct)
      {
        if ((!strcmp(pAct->szMnemonic,"call")) && (!strcmp(pAct->szOperands,"memset")))
        {
          pMod = pAct;
          pMod = pMod->pPrev;if (!pMod) goto AbortPP;
          pMod = pMod->pPrev;if (!pMod) goto AbortPP;
          pMod = pMod->pPrev;if (!pMod) goto AbortPP;
          sprintf(szOpr,"%s,%s",pMod->szOperands,pMod->szOperands);
          InsertMnemonic(pMod,"add",szOpr);
          pAct = pAct->pNext;
          break;
        }
        pAct = pAct->pNext;
      }
      if (!pAct) goto AbortPP;
      while (pAct)
      {
        if ((!strcmp(pAct->szMnemonic,"call")) && (!strcmp(pAct->szOperands,"memset")))
        {
          pMod = pAct;
          pMod = pMod->pPrev;if (!pMod) goto AbortPP;
          pMod = pMod->pPrev;if (!pMod) goto AbortPP;
          pMod = pMod->pPrev;if (!pMod) goto AbortPP;
          sprintf(szOpr,"%s,%s",pMod->szOperands,pMod->szOperands);
          InsertMnemonic(pMod,"add",szOpr);
          pAct = pAct->pNext;
          break;
        }
        pAct = pAct->pNext;
      }
      if (!pAct) goto AbortPP;
      pAct=pAct->pNext;if (!pAct) goto AbortPP;
      pAct=pAct->pNext;if (!pAct) goto AbortPP;
      pAct=pAct->pNext;if (!pAct) goto AbortPP;
      pAct->szMnemonic[0]=0;
      pAct=pAct->pNext;if (!pAct) goto AbortPP;
      pAct=pAct->pNext;if (!pAct) goto AbortPP;
      cnt=0;
      while (pAct)
      {
        if ((!strcmp(pAct->szMnemonic,"lea")) && (!strcmp(pAct->szOperands,"(%edi,%edi,1),%ecx")))
        {   
          cnt++;
          if (cnt==2)
          {                                                                 
            break;
          }
        }
        pAct->szMnemonic[0]=0;
        pAct = pAct->pNext;
      }
      if (!pAct) goto AbortPP;

      while (pAct)
      {
        if ((!strcmp(pAct->szMnemonic,"movl")) && (NULL!=(p=strstr(pAct->szOperands,"$0x0,0x"))))
        {                                                                    
          addr_uFileOffset = strtoul(p+7,&p,16);
          if (!addr_uFileOffset) goto AbortPP;
          break;
        }
        pAct = pAct->pNext;
      }
      if (!pAct) goto AbortPP;
      break;

    case 23:  // pgopti_Dump_Segment
      pAct = pRoot;
      while (pAct)
      {
        if ((!strcmp(pAct->szMnemonic,"call")) && (!strcmp(pAct->szOperands,"fprintf")))
        {
          pAct->szMnemonic[0]=0;
          pMod = pAct->pPrev;if (!pMod) goto AbortPP;
          pMod->szMnemonic[0]=0;
          pMod = pMod->pPrev;if (!pMod) goto AbortPP;
          pMod->szMnemonic[0]=0;
          pAct=pAct->pNext;if (!pAct) goto AbortPP;
          pAct->szMnemonic[0]=0;
        }
        pAct = pAct->pNext;
      }
      break;

    case 26:  // pgopti_Sizeof_Segment
      pAct = pRoot;
      while (pAct)
      {
        if ((!strcmp(pAct->szMnemonic,"call")) && (!strcmp(pAct->szOperands,"abort")))
        {
          strcpy(pAct->szMnemonic,"add");
          strcpy(pAct->szOperands,"$0x14,%esp");
          pAct=pAct->pNext;
          if (!pAct) goto AbortPP;
          InsertMnemonic(pAct,"pop","%ebx");
          InsertMnemonic(pAct,"pop","%ebp");
          InsertMnemonic(pAct,"pop","%esi");
          InsertMnemonic(pAct,"pop","%edi");
          InsertMnemonic(pAct,"sub","%eax,%eax");
          InsertMnemonic(pAct,"ret","");
          break;
        }
        pAct = pAct->pNext;
      }
      if (!pAct) goto AbortPP;
      break;

    case 27:  // pgopti_Prof_Dump_In_Progress
      pAct = pRoot;
      while (pAct)
      {
        if ((!strcmp(pAct->szMnemonic,"call")) && (!strcmp(pAct->szOperands,"InterlockedCompareExchangePointer")))
        {
          pMod = pAct->pPrev;if (!pMod) goto AbortPP;
          if ((strcmp(pMod->szMnemonic,"push")) || (pMod->szOperands[0]!='$')) goto AbortPP;
          addr_uLock = strtoul(&pMod->szOperands[1],&p,0);
          if (!addr_uLock) goto AbortPP;
          break;
        }
        pAct = pAct->pNext;
      }
      if (!pAct) goto AbortPP;
      break;

    case 31:  // pgopti_Value_Prof_Clear_Non_Steady
      pAct = pRoot;
      while (pAct)
      {
        if ((!strcmp(pAct->szMnemonic,"pushl")) && (NULL!=(p=strstr(pAct->szOperands,"0x"))))
        {
          addr_uMaxVal = strtoul(p+2,&p,16);
          if (!addr_uMaxVal) goto AbortPP;
          break;
        }
        pAct = pAct->pNext;
      }
      if (!pAct) goto AbortPP;
      break;
  }

  return 0;
}

int instructions ( PFLINE pRun )
{
  while ((pRun) && (!pRun->szMnemonic[0])) pRun=pRun->pNext;
  return (!pRun) ? 0 : 1;
}

int main ( int argc, char *argv[] )
{
  char                szFunction[256],szFile[256];
  int                 i,l,err,j;
  FILE               *hDst;
  PFLINE              pRun;
  char               *p,*p2,*pLIC;
  unsigned long       uVal;
  int                 iSize=0;

  if (argc!=3)
    return 1;

  memset(pFunctions,0,sizeof(pFunctions));
  memset(uLabAddrs,0,sizeof(uLabAddrs));
  memset(iLabRefs,0,sizeof(iLabRefs));

  szLine[0]=0;
  hFile = popen("which icc","r");
  if (!hFile)
  {
    fprintf(stderr,"pipe/fork error.\n");
    return 1;
  }
  fgets(szLine,sizeof(szLine),hFile);
  while (fgets(szCopyright,sizeof(szCopyright),hFile)) {}
  if (-1==pclose(hFile))
  {
    fprintf(stderr,"pipe/fork error.\n");
    return 1;
  }
  p=strstr(szLine,"/bin/icc");
  if (!p)
  {
    fprintf(stderr,"error: cannot find 'icc'.\n");
    return 1;
  }
  *p=0;
  strcat(szLine,"/licenses");
  sprintf(szCopyright,"find %s -name *.lic -print",szLine);
  szLine[0]=0;
  hFile = popen(szCopyright,"r");
  if (!hFile)
  {
    fprintf(stderr,"pipe/fork error.\n");
    return 1;
  }
  fgets(szLine,sizeof(szLine),hFile);
  while (fgets(szCopyright,sizeof(szCopyright),hFile)) {}
  if (-1==pclose(hFile))
  {
    fprintf(stderr,"pipe/fork error.\n");
    return 1;
  }
  p=strstr(szLine,".lic");
  if (!p)
  {
    fprintf(stderr,"error: cannot find icc license.\n");
    return 1;
  }
  l=strlen(szLine);
  if ((l) && (szLine[l-1]==0x0a)) szLine[l-1]=0;
  fprintf(stdout,"'%s'\n",szLine);
  hFile = fopen(szLine,"rb");
  if (!hFile)
  {
    fprintf(stderr,"error: cannot read icc license.\n");
    return 1;
  }
  fseek(hFile,0L,SEEK_END);
  iSize = ftell(hFile);
  fseek(hFile,0L,SEEK_SET);
  if (!iSize)
  {
    ReadError:
    fclose(hFile);
    fprintf(stderr,"error: cannot read icc license.\n");
    return 1;
  }
  pLIC = (char *)malloc(iSize);
  if (!pLIC) goto ReadError;
  if (fread(pLIC,1,iSize,hFile)!=iSize) goto ReadError;
  fclose(hFile);

  memset(szCopyright,0,sizeof(szCopyright));
  hFile = fopen(argv[1],"rt");
  if (!hFile)
    return 1;

  while (fgets(szLine,sizeof(szLine),hFile))
  {
    l=strlen(szLine);
    p=IgnWS(szLine);
    if (!(*p))
    {
      while (fgets(szLine,sizeof(szLine),hFile)) {}
      break;
    }
    strcat(szCopyright,szLine);
  }
  fclose(hFile);

  hFile = fopen(argv[2],"rt");
  if (!hFile) return 1;

  szLine[0]=0;
  while (fgets(szLine,sizeof(szLine),hFile))
  {
    l=strlen(szLine);
    if ((ISHEX(szLine[0])) && (ISHEX(szLine[1])) &&
        (ISHEX(szLine[2])) && (ISHEX(szLine[3])) &&
        (ISHEX(szLine[4])) && (ISHEX(szLine[5])) &&
        (ISHEX(szLine[6])) && (ISHEX(szLine[7])) &&
        (szLine[8]==' ') && (szLine[9]=='<') && 
        (l>=3) && (szLine[l-3]=='>') && (szLine[l-2]==':'))
    {
      memcpy(szFunction,&szLine[10],l-13);
      szFunction[l-13]=0;
      if (ProcessFunction(szFunction))
      {
        fclose(hFile);
        return 1;
      }
    }
    szLine[0]=0;
  }
  fclose(hFile);

  err=0;
  for (i=0;i<labels;i++)
    if (!iLabRefs[i])
    {
      err=1;
      fprintf(stderr,"no reference to label _L%03lX.\n",i);
    }
  if (err) return 1;

  ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  hFile = fopen("template_intlPGO.S","rt");
  hDst  = popen("as -o intlPGO.object","w");

  if ((!hFile) || (!hDst))
  {
    if (hFile) fclose(hFile);
    if (hDst) pclose(hDst);
    fprintf(stderr,"File/pipe error.\n");
    return 1;
  }

  while (fgets(szLine,sizeof(szLine),hFile))
    fprintf(hDst,"%s",szLine);
  fclose(hFile);

  for (i=0;i<NUM_FUNC;i++)
  {
    if (!pFunctions[i])
    {
      fprintf(stderr,"NULL pointer (%s).\n",Functions[i].szFunction);
      pclose(hDst);
      return 1;
    }
    if (Functions[i].grep)
    {
      if (Functions[i].align<0)
      {
        fprintf(hDst,"\t.globl\t\t%s\n",Functions[i].szFunction);
        Functions[i].align = -Functions[i].align;
      }
      fprintf(hDst,"\t.p2align\t\t%i\n",Functions[i].align);
      fprintf(hDst,"%s:\n",Functions[i].szFunction);
      pRun = pFunctions[i];

      while (pRun)
      {
        for (j=0;j<labels;j++) if (uLabAddrs[j]==pRun->uAddr) break;
        if (j<labels) fprintf(hDst,"_L%03lX:\n",j);
        if (pRun->szMnemonic[0])
        {
          fprintf(hDst,"\t%s\t\t",pRun->szMnemonic);
          if (!strcmp(pRun->szOperands,"memset")) fprintf(hDst,"__memset__");
          else
          if (!strcmp(pRun->szOperands,"strlen")) fprintf(hDst,"__strlen__");
          else
          if ((pRun->szOperands[0]=='_') && (pRun->szOperands[1]=='L') && 
             (ISHEX(pRun->szOperands[2])) && (ISHEX(pRun->szOperands[3])) && (ISHEX(pRun->szOperands[4])))
          {
            uVal = strtoul(&pRun->szOperands[2],&p,16);
            if (uVal>=((unsigned long)labels))
            {
              fprintf(stderr,"jump to undefined label detected.\n");
              pclose(hDst);
              return 1;
            }
            if (!pRun->uAddr)
            {
              fprintf(stderr,"cannot determine forward/backward reference.\n");
              pclose(hDst);
              return 1;
            }
            if (uLabAddrs[uVal]<=pRun->uAddr)
              fprintf(hDst,"%s",pRun->szOperands);
            else
              fprintf(hDst,"%s",pRun->szOperands);
          }
          else
          {
            p=pRun->szOperands;
            while (p2=(strstr(p,"0x")))
            {
              p2[0]=0;
              fprintf(hDst,"%s",p);
              p2[0]='0';
              uVal = strtoul(p2+2,&p,16);
              if (uVal==addr_is64bit) fprintf(hDst,"is64bit");
              else
              if (uVal==addr_uValue1) fprintf(hDst,"uValue1");
              else
              if (uVal==addr_uValue2) fprintf(hDst,"uValue2");
              else
              if (uVal==addr_pSegments) fprintf(hDst,"pSegments");
              else
              if (uVal==addr_uFileOffset) fprintf(hDst,"uFileOffset");
              else
              if (uVal==addr_uMaxVal) fprintf(hDst,"uMaxVal");
              else
              if (uVal==addr_uLock) fprintf(hDst,"uLock");
              else
              fprintf(hDst,"0x%lx",uVal);
            }
            if (!p2) fprintf(hDst,"%s",p);
          }

          if (!pRun->uBytes)
            fprintf(hDst,"\t\t\t\t# THIS LINE WAS MODIFIED (INSERTED).\n");
          else
            fprintf(hDst,"\n");
        }
        else
          fprintf(hDst,"# THIS LINE WAS MODIFIED (DELETED).\n");

        if (((!strcmp(pRun->szMnemonic,"ret")) || (!strcmp(pRun->szMnemonic,"jmp"))) && (instructions(pRun->pNext)))
          fprintf(hDst,"\t.p2align\t\t2\n");

        pRun = pRun->pNext;
      }
      fprintf(hDst,"\t.type\t\t%s,@function\n",Functions[i].szFunction);
      fprintf(hDst,"\t.size\t\t%s,.-%s\n\n",Functions[i].szFunction,Functions[i].szFunction);
    }
  }

  fprintf(hDst,"\n\t.data\n");
  fprintf(hDst,"\t.globl\tszICCCopyright\n");
  fprintf(hDst,"\t.align\t\t4,0x00\n");
  fprintf(hDst,"szICCCopyright:\n");
  i=0;
  while (szCopyright[i])
  {
    fprintf(hDst,"\t.byte\t\t0x%02lx\n",((unsigned long)szCopyright[i])&0xff);
    i++;
  }
  fprintf(hDst,"\t.byte\t\t0x0A\n\n");
  fprintf(hDst,"\t.byte\t\t0x00\n\n");
  fprintf(hDst,"\t.type\t\tszICCCopyright,@object\n");
  fprintf(hDst,"\t.size\t\tszICCCopyright,%i\n\n",i+2);

  fprintf(hDst,"\t.globl\tszICCLicense\n");
  fprintf(hDst,"\t.align\t\t4,0x00\n");
  fprintf(hDst,"szICCLicense:\n");
  for (i=0;i<iSize;i++) fprintf(hDst,"\t.byte\t\t0x%02lx\n",((unsigned long)pLIC[i])&0xff);
  fprintf(hDst,"\t.byte\t\t0x0A\n\n");
  fprintf(hDst,"\t.byte\t\t0x00\n\n");
  fprintf(hDst,"\t.type\t\tszICCLicense,@object\n");
  fprintf(hDst,"\t.size\t\tszICCLicense,%i\n\n",iSize+2);

  fprintf(hDst,"# End\n");
  if (-1==pclose(hDst))
  {
    fprintf(stderr,"ERROR: could not create 'intlPGO.object'.\n");
    return 1;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  hFile = fopen("intlPGO.object","rb");
  if (!hFile)
  {
    fprintf(stderr,"ERROR: could not create 'intlPGO.object'.\n");
    return 1;
  }
  fclose(hFile);

  return 0;
}














